from sapextractor.algo.p2p import p2p_common
from sapextractor.utils.graph_building import build_graph
from sapextractor.utils.filters import case_filter
from sapextractor.utils import constants


def apply(con, ref_type="EKKO"):
    dataframe, G, nodes_types = p2p_common.extract_tables_and_graph(con)
    dataframe = dataframe[[x for x in dataframe.columns if x.startswith("event_")]]
    anc_succ = build_graph.get_ancestors_successors_from_graph(G, nodes_types, ref_type=ref_type)
    dataframe = dataframe.merge(anc_succ, left_on="event_node", right_on="node", suffixes=('', '_r'), how="right")
    dataframe = dataframe.dropna(subset=["event_activity", "event_timestamp"])
    cols = {x: x.split("event_")[-1] for x in dataframe.columns}
    cols["event_activity"] = "concept:name"
    cols["event_timestamp"] = "time:timestamp"
    cols["event_USERNAME"] = "org:resource"
    dataframe = dataframe.rename(columns=cols)
    dataframe = dataframe.sort_values("time:timestamp")
    dataframe = dataframe.loc[:,~dataframe.columns.duplicated()]
    dataframe = case_filter.filter_on_case_size(dataframe, "case:concept:name", min_case_size=1, max_case_size=constants.MAX_CASE_SIZE)
    return dataframe


def cli(con):
    print("\n\nP2P - XES log\n")
    ref_type = input("Provide the central table for the extraction (default: EKKO):")
    if not ref_type:
        ref_type = "EKKO"
    dataframe = apply(con, ref_type=ref_type)
    path = input("Insert the path where the log should be saved (default: p2p.csv): ")
    if not path:
        path = "p2p.csv"
    dataframe.to_csv(path, sep=",", quotechar="\"",  index=False)



import pandas as pd
from sapextractor.utils.blart import extract_blart


def apply(con, gjahr="2020"):
    bkpf = con.prepare_and_execute_query("BKPF", ["BELNR", "GJAHR", "AWKEY", "BLART"], additional_query_part=" WHERE GJAHR = '"+gjahr+"'")
    blart_vals = set(bkpf["BLART"].unique())
    blart_vals = {x: x for x in blart_vals}
    blart_vals = extract_blart.apply_static(con, doc_types=blart_vals)
    bkpf = bkpf.to_dict("records")
    try:
        bseg = con.prepare_and_execute_query("BSAK", ["BELNR", "BUZEI", "AUGDT", "AUGBL"])
    except:
        bseg = con.prepare_and_execute_query("BSEG", ["BELNR", "BUZEI", "AUGDT", "AUGBL"])
    bseg = bseg.dropna(subset=["AUGBL"])
    try:
        bseg["AUGDT"] = pd.to_datetime(bseg["AUGDT"], format="%d.%m.%Y")
    except:
        bseg["AUGDT"] = pd.to_datetime(bseg["AUGDT"])
    bseg = bseg.to_dict("records")
    dict_awkey = {}
    clearance_docs_dates = {}
    blart_dict = {}
    for el in bkpf:
        awkey = el["AWKEY"]
        belnr = el["BELNR"]
        blart = el["BLART"]
        if awkey is not None:
            if len(awkey) == 18:
                awkey = awkey[4:-4]
                if awkey not in dict_awkey:
                    dict_awkey[awkey] = []
                dict_awkey[awkey].append(belnr)
        if not belnr in blart_dict:
            blart_dict[belnr] = blart
    for el in bseg:
        belnr = el["BELNR"]
        augbl = el["AUGBL"]
        augdt = el["AUGDT"]
        if augbl in blart_dict:
            blart = blart_dict[augbl]
            if not belnr in clearance_docs_dates:
                clearance_docs_dates[belnr] = set()
            clearance_docs_dates[belnr].add((augbl, augdt, blart))

    return dict_awkey, clearance_docs_dates, blart_vals



from sapextractor.algo.p2p import p2p_1d_dataframe
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.log.util import sorting
from pm4py.objects.log.exporter.xes import exporter as xes_exporter


def apply(con, ref_type="EKKO"):
    dataframe = p2p_1d_dataframe.apply(con, ref_type=ref_type)
    log = log_converter.apply(dataframe, parameters={"stream_postprocessing": True})
    log = sorting.sort_timestamp(log, "time:timestamp")
    return log


def cli(con):
    print("\n\nP2P - XES log\n")
    ref_type = input("Provide the central table for the extraction (default: EKKO):")
    if not ref_type:
        ref_type = "EKKO"
    log = apply(con, ref_type=ref_type)
    path = input("Insert the path where the log should be saved (default: p2p.xes): ")
    if not path:
        path = "p2p.xes"
    xes_exporter.apply(log, path)



from sapextractor.algo.ap_ar import ap_ar_common
from sapextractor.utils.filters import case_filter
from sapextractor.utils import constants


def apply(con, gjahr="1997", bukrs=None, **ext_arg):
    bkpf = ap_ar_common.get_full_dataframe(con, gjahr=gjahr, bukrs=bukrs, filter_columns=True)
    bkpf["event_activity"] = bkpf["event_ONLYACT"]
    ren_cols = {"event_activity": "concept:name", "event_timestamp": "time:timestamp", "event_USNAM": "org:resource"}
    bkpf = bkpf.rename(columns=ren_cols)
    ren_cols = {x: x.split("event_")[-1] for x in bkpf.columns}
    bkpf = bkpf.rename(columns=ren_cols)
    bkpf["case:concept:name"] = bkpf["BELNR"]
    bkpf = bkpf.dropna(subset=["concept:name", "time:timestamp"], how="any")
    bkpf = bkpf.sort_values("time:timestamp")
    bkpf = case_filter.filter_on_case_size(bkpf, "case:concept:name", min_case_size=1, max_case_size=constants.MAX_CASE_SIZE)
    return bkpf


def cli(con):
    print("\n\nAccounting - Transactions for the single document (dataframe)\n")
    dataframe = apply(con)
    path = input("Insert the path where the dataframe should be saved (default: bkpf.csv):")
    if not path:
        path = "bkpf.csv"
    dataframe.to_csv(path, sep=",", quotechar="\"", index=False)



import numpy as np
import pandas as pd

from sapextractor.algo.o2c import o2c_common
from sapextractor.utils import constants
from sapextractor.utils.change_tables import extract_change
from sapextractor.utils.filters import case_filter
from sapextractor.utils.graph_building import build_graph
from sapextractor.algo.o2c import payment_part


def extract_changes_vbfa(con, dataframe):
    case_vbeln = dataframe[["case:concept:name", "VBELN"]].to_dict("records")
    case_vbeln_dict = {}
    for x in case_vbeln:
        caseid = x["case:concept:name"]
        vbeln = x["VBELN"]
        if vbeln not in case_vbeln_dict:
            case_vbeln_dict[vbeln] = set()
        case_vbeln_dict[vbeln].add(caseid)
    ret = []
    for tup in [("VERKBELEG", "VBAK"), ("VERKBELEG", "VBAP"), ("VERKBELEG", "VBUK"), ("LIEFERUNG", "LIKP"),
                ("LIEFERUNG", "LIPS"), ("LIEFERUNG", "VBUK")]:
        changes = extract_change.apply(con, objectclas=tup[0], tabname=tup[1])
        changes = {x: y for x, y in changes.items() if x in case_vbeln_dict}
        for x, y in changes.items():
            y = y[[xx for xx in y.columns if xx.startswith("event_")]]
            cols = {x: x.split("event_")[-1] for x in y.columns}
            cols["event_timestamp"] = "time:timestamp"
            y = y.rename(columns=cols)
            y["VBELN"] = y["AWKEY"]
            y["concept:name"] = y["CHANGEDESC"]
            for cc in case_vbeln_dict[x]:
                z = y.copy()
                z["case:concept:name"] = cc
                ret.append(z)

    if ret:
        ret = pd.concat(ret)
    else:
        ret = pd.DataFrame()

    return ret


def extract_bkpf_bsak(con, dataframe, gjahr="2020"):
    case_vbeln = dataframe[["case:concept:name", "VBELN"]].to_dict("records")
    case_vbeln_dict = {}
    for x in case_vbeln:
        caseid = x["case:concept:name"]
        vbeln = x["VBELN"]
        if vbeln not in case_vbeln_dict:
            case_vbeln_dict[vbeln] = set()
        case_vbeln_dict[vbeln].add(caseid)

    dict_awkey, clearance_docs_dates, blart_vals = payment_part.apply(con, gjahr=gjahr)

    intersect = set(case_vbeln_dict.keys()).intersection(dict_awkey.keys())

    ret = []
    for k in intersect:
        for belnr in dict_awkey[k]:
            if belnr in clearance_docs_dates:
                for clearingdoc in clearance_docs_dates[belnr]:
                    for cas in case_vbeln_dict[k]:
                        ret.append(
                            {"case:concept:name": cas, "concept:name": "Clearance (" + blart_vals[clearingdoc[2]] + ")",
                             "AUGBL": clearingdoc[0], "time:timestamp": clearingdoc[1]})
    ret = pd.DataFrame(ret)

    if len(ret) > 0:
        if "time:timestamp" in ret.columns:
            ret["time:timestamp"] = ret["time:timestamp"] + pd.Timedelta(np.timedelta64(86399, 's'))

            ret = ret.groupby(["case:concept:name", "AUGBL"]).first().reset_index()

    return ret


def apply(con, ref_type="Order", keep_first=True, min_extr_date="2020-01-01 00:00:00", gjahr="2020", enable_changes=True, enable_payments=True, allowed_act_doc_types=None, allowed_act_changes=None):
    dataframe = o2c_common.apply(con, keep_first=keep_first, min_extr_date=min_extr_date)
    dataframe = dataframe[[x for x in dataframe.columns if x.startswith("event_")]]
    cols = {x: x.split("event_")[-1] for x in dataframe.columns}
    cols["event_activity"] = "concept:name"
    cols["event_timestamp"] = "time:timestamp"
    dataframe = dataframe.rename(columns=cols)
    ancest_succ = build_graph.get_ancestors_successors(dataframe, "VBELV", "VBELN", "VBTYP_V", "VBTYP_N",
                                                       ref_type=ref_type)
    # ancest_succ = build_graph.get_conn_comp(dataframe, "VBELV", "VBELN", "VBTYP_V", "VBTYP_N", ref_type=ref_type)
    dataframe = dataframe.merge(ancest_succ, left_on="VBELN", right_on="node", suffixes=('', '_r'), how="right")
    dataframe = dataframe.reset_index()
    if keep_first:
        dataframe = dataframe.groupby(["case:concept:name", "VBELN"]).first().reset_index()
    if allowed_act_doc_types is not None:
        allowed_act_doc_types = set(allowed_act_doc_types)
        dataframe = dataframe[dataframe["concept:name"].isin(allowed_act_doc_types)]
    if enable_changes:
        changes = extract_changes_vbfa(con, dataframe)
    else:
        changes = pd.DataFrame()
    if enable_payments:
        payments = extract_bkpf_bsak(con, dataframe, gjahr=gjahr)
    else:
        payments = pd.DataFrame()
    if allowed_act_changes is not None:
        allowed_act_changes = set(allowed_act_changes)
        changes = changes[changes["concept:name"].isin(allowed_act_changes)]
    dataframe = pd.concat([dataframe, changes, payments])
    dataframe = dataframe.sort_values("time:timestamp")
    dataframe = case_filter.filter_on_case_size(dataframe, "case:concept:name", min_case_size=1,
                                                max_case_size=constants.MAX_CASE_SIZE)
    return dataframe


def cli(con):
    print("\n\nO2C dataframe extractor\n")
    ref_type = input("Insert the central document type of the extraction (default: Invoice): ")
    if not ref_type:
        ref_type = "Invoice"
    ext_type = input("Do you want to extract the document log, or the items log (default: document):")
    if not ext_type:
        ext_type = "document"
    keep_first = True
    if ext_type == "document":
        keep_first = True
    elif ext_type == "items":
        keep_first = False
    min_extr_date = input("Insert the minimum extraction date (default: 2020-01-01 00:00:00): ")
    if not min_extr_date:
        min_extr_date = "2020-01-01 00:00:00"
    gjahr = input("Insert the fiscal year (default: 2020):")
    if not gjahr:
        gjahr = "2020"
    dataframe = apply(con, ref_type=ref_type, keep_first=keep_first, min_extr_date=min_extr_date, gjahr=gjahr)
    path = input("Insert the path where the dataframe should be saved (default: o2c.csv):")
    if not path:
        path = "o2c.csv"
    dataframe.to_csv(path, sep=",", quotechar="\"", index=False)



from sapextractor.utils.tstct import extract_tstct
from sapextractor.utils.dates import timestamp_column_from_dt_tm


def apply(con):
    mkpf = con.prepare_and_execute_query("MKPF", ["MBLNR", "BLART", "CPUDT", "CPUTM", "USNAM", "TCODE"])
    mkpf.columns = ["event_"+x for x in mkpf.columns]
    mkpf = mkpf.rename(columns={"event_USNAM": "event_USERNAME"})
    mkpf = mkpf.dropna(subset=["event_MBLNR", "event_TCODE"], how="any")
    transactions = set(mkpf["event_TCODE"].unique())
    tcodes = extract_tstct.apply_static(con, transactions=transactions)
    mkpf["event_activity"] = mkpf["event_TCODE"].map(tcodes)
    mkpf = timestamp_column_from_dt_tm.apply(mkpf, "event_CPUDT", "event_CPUTM", "event_timestamp")
    mkpf = mkpf.dropna(subset=["event_timestamp"], how="any")
    mkpf["event_FROMTABLE"] = "MKPF"
    mkpf["event_node"] = "MKPF_" + mkpf["event_MBLNR"]
    mkpf_node_types = {x: "MKPF" for x in mkpf["event_node"].unique()}
    return mkpf, mkpf_node_types



from dateutil import parser


def perform_mapping(tabname, fname, value_old, value_new, chngind, fnames, return_however=True):
    if fname not in fnames or fnames[fname] is None:
        fnames[fname] = fname
    if fname == "KOSTK" and value_new == "B":
        return "Picking: Partially Processed", 1
    elif fname == "KOSTK" and value_new == "C":
        return "Picking: Completely Processed", 1
    elif fname == "GBSTK" and value_new == "B":
        return "Set Order Status: Partially Processed", 1
    elif fname == "GBSTK" and value_new == "C":
        return "Set Order Status: Completely Processed", 1
    elif fname == "KOQUK" and value_new == "B":
        return "Pick Confirmation: Partially Processed", 1
    elif fname == "KOQUK" and value_new == "C":
        return "Pick Confirmation: Completely Processed", 1
    elif fname == "LVSTK" and value_new == "B":
        return "Warehouse Management: Partially Processed", 1
    elif fname == "LVSTK" and value_new == "C":
        return "Warehouse Management: Completely Processed", 1
    elif fname == "WBSTK" and value_new == "B":
        return "Total Goods Movement: Partially processed", 1
    elif fname == "WBSTK" and value_new == "C":
        return "Total Goods Movement: Completely processed", 1
    elif fname == "TRSTA" and value_new == "B":
        return "Transportation Status: Partially processed", 1
    elif fname == "TRSTA" and value_new == "C":
        return "Transportation Status: Completely processed", 1
    elif fname == "SPE_IMWRK" and value_new == "X":
        return "Item in Plant", 1
    elif fname == "MPROK" and value_new == "A":
        return "Manual price change carried out", 1
    elif fname == "MPROK" and value_new == "B":
        return "Condition manually deleted", 1
    elif fname == "MPROK" and value_new == "C":
        return "Manual price change released", 1
    elif fname == "SPE_REV_VLSTK" and value_new == "A":
        return "Distribution Status: Relevant", 1
    elif fname == "SPE_REV_VLSTK" and value_new == "B":
        return "Distribution Status: Distributed", 1
    elif fname == "SPE_REV_VLSTK" and value_new == "C":
        return "Distribution Status: Confirmed", 1
    elif fname == "SPE_REV_VLSTK" and value_new == "D":
        return "Distribution Status: Planned for Distribution", 1
    elif fname == "SPE_REV_VLSTK" and value_new == "E":
        return "Distribution Status: Delivery split was performed locally", 1
    elif fname == "SPE_REV_VLSTK" and value_new == "F":
        return "Distribution Status: Change Management Switched Off", 1
    elif fname == "VLSTK" and value_new == "A":
        return "Distribution Status: Relevant", 1
    elif fname == "VLSTK" and value_new == "B":
        return "Distribution Status: Distributed", 1
    elif fname == "VLSTK" and value_new == "C":
        return "Distribution Status: Confirmed", 1
    elif fname == "VLSTK" and value_new == "D":
        return "Distribution Status: Planned for Distribution", 1
    elif fname == "VLSTK" and value_new == "E":
        return "Distribution Status: Delivery split was performed locally", 1
    elif fname == "VLSTK" and value_new == "F":
        return "Distribution Status: Change Management Switched Off", 1
    elif fname == "SPE_GEN_ELIKZ" and value_new == "X":
        return "Delivery Completed", 1
    elif fname == "IMWRK" and value_new == "X":
        return "Delivery in Plant", 1
    elif fname == "UVVLS" and value_new == "A":
        return "Delivery Uncompletion: Not yet processed", 1
    elif fname == "UVVLS" and value_new == "B":
        return "Delivery Uncompletion: Partially processed", 1
    elif fname == "UVVLS" and value_new == "C":
        return "Delivery Uncompletion: Completely processed", 1
    elif fname == "FKDAT":
        try:
            value_new = parser.parse(value_new).timestamp()
        except:
            value_new = 0
        try:
            value_old = parser.parse(value_old).timestamp()
        except:
            value_old = 0
        if value_old == 0:
            return "Set Billing Date", 2
        elif value_new == 0:
            return "Remove Billing Date", 2
        elif value_new <= value_old:
            return "Anticipate Billing Date", 1
        else:
            return "Postpone Billing Date", 1
    elif fname == "FAKSK":
        if value_new is None:
            return "Remove Billing Block", 2
        else:
            return "Set Billing Block", 2
    elif fname == "SKFBP":
        value_new = float(value_new)
        value_old = float(value_old)
        if value_new <= value_old:
            return "Decrease Cash Discount", 1
        else:
            return "Increase Cash Discount", 1
    elif fname == "UVALL" and value_new == "A":
        return "Order Uncompletion: Not yet processed", 1
    elif fname == "UVALL" and value_new == "B":
        return "Order Uncompletion: Partially processed", 1
    elif fname == "UVALL" and value_new == "C":
        return "Order Uncompletion: Completely processed", 1
    elif fname == "LISPL" and value_new == "A":
        return "Delivery is for single warehouse", 1
    elif fname == "LISPL" and value_new == "B":
        return "Delivery is not for single warehouse", 1
    elif fname == "WADAT_IST":
        try:
            value_new = parser.parse(value_new).timestamp()
        except:
            value_new = 0
        try:
            value_old = parser.parse(value_old).timestamp()
        except:
            value_old = 0
        if value_old == 0:
            return "Set Actual Goods Movement Date", 2
        elif value_new == 0:
            return "Remove Actual Goods Movement Date", 2
        elif value_new <= value_old:
            return "Anticipate Actual Goods Movement Date", 1
        else:
            return "Postpone Actual Goods Movement Date", 1
    elif fname == "KEY" and tabname == "FPLT" and chngind == "I":
        return "Insert Billing Plan", 2
    elif fname == "KEY" and tabname == "FPLT" and chngind == "D":
        return "Remove Billing Plan", 2
    elif fname == "KEY" and tabname == "VBEP" and chngind == "D":
        return "Remove Schedule Line", 2
    elif fname == "KEY" and tabname == "VBEP" and chngind == "I":
        return "Insert Schedule Line", 2
    elif fname == "KEY" and tabname == "VBAK" and chngind == "D":
        return "Cancel Order", 2
    elif fname == "KEY" and tabname == "VBAP" and chngind == "D":
        return "Remove Order Item", 2
    elif fname == "KEY" and tabname == "VBAP" and chngind == "I":
        return "Insert Order Item", 2
    elif fname == "KEY" and tabname == "LIPS" and chngind == "D":
        return "Remove Delivery Item", 2
    elif fname == "KEY" and tabname == "LIPS" and chngind == "I":
        return "Insert Delivery Item", 2
    else:
        if return_however:
            return "Changed "+fnames[fname], 0
    return None, None




from sapextractor.utils.change_tables import mapping


def extract(con):
    df = con.execute_read_sql(
                              "SELECT OBJECTCLAS, TABNAME, FNAME, VALUE_OLD, VALUE_NEW, CHNGIND, Count(*) AS COUNT FROM (SELECT OBJECTCLAS, TABNAME, FNAME, VALUE_OLD, VALUE_NEW, CHNGIND FROM "+con.table_prefix+"CDPOS WHERE OBJECTCLAS = 'VERKBELEG' OR OBJECTCLAS = 'LIEFERUNG') GROUP BY OBJECTCLAS, TABNAME, FNAME, VALUE_OLD, VALUE_NEW, CHNGIND", ["OBJECTCLAS", "TABNAME", "FNAME", "VALUE_OLD", "VALUE_NEW", "CHNGIND", "COUNT"])
    stream = df.to_dict("r")
    fnames = {}
    counter = {}
    i = 0
    while i < len(stream):
        el = stream[i]

        change, typ = mapping.perform_mapping(el["TABNAME"], el["FNAME"], el["VALUE_OLD"], el["VALUE_NEW"], el["CHNGIND"], fnames, return_however=False)
        el["CHANGEDESC"] = (change, typ, el["OBJECTCLAS"])
        if el["CHANGEDESC"][0] is None:
            del stream[i]
            continue
        if not el["CHANGEDESC"] in counter:
            counter[el["CHANGEDESC"]] = 0

        counter[el["CHANGEDESC"]] += int(el["COUNT"])
        i = i + 1

    counter = [[x, y] for x, y in counter.items()]
    counter = sorted(counter, key=lambda x: x[1], reverse=True)

    return counter



from sapextractor.algo.o2c import o2c_1d_dataframe_extractor, o2c_1d_log_extractor
from sapextractor.algo.o2c import obj_centr_log


def cli(con):
    print("\n\nO2C extraction\n")
    print("available extraction types:")
    print("1) O2C classic log (XES log)")
    print("2) O2C classic log (dataframe)")
    print("3) O2C object-centric log")
    print()
    ext_type = input("insert your choice (default: 1): ")
    if not ext_type:
        ext_type = "1"
    if ext_type == "1":
        return o2c_1d_log_extractor.cli(con)
    elif ext_type == "2":
        return o2c_1d_dataframe_extractor.cli(con)
    elif ext_type == "3":
        return obj_centr_log.cli(con)



import pandas as pd


def apply(con):
    ekko = con.prepare_and_execute_query("EKKO", ["EBELN", "ERNAM", "AEDAT", "LIFNR", "ZTERM"])
    ekko.columns = ["event_"+x for x in ekko.columns]
    ekko = ekko.rename(columns={"event_ERNAM": "event_USERNAME", "event_AEDAT": "event_timestamp"})
    ekko = ekko.sort_values("event_timestamp")
    ekko = ekko.groupby("event_EBELN").first().reset_index()
    ekko["event_timestamp"] = pd.to_datetime(ekko["event_timestamp"])
    ekko["event_activity"] = "Create Purchase Order"
    ekko["event_FROMTABLE"] = "EKKO"
    ekko["event_node"] = "EKKO_"+ekko["event_EBELN"]
    ekko = ekko.dropna(subset=["event_node"], how="any")
    ekko_nodes_types = {x: "EKKO" for x in ekko["event_node"].unique()}
    return ekko, ekko_nodes_types



from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.log.exporter.xes import exporter as xes_exporter
from pm4py.objects.log.util import sorting

from sapextractor.algo.o2c import o2c_1d_dataframe_extractor


def apply(con, ref_type="Invoice", keep_first=True, min_extr_date="2020-01-01 00:00:00", gjahr="2020", enable_changes=True, enable_payments=True, allowed_act_doc_types=None, allowed_act_changes=None):
    dataframe = o2c_1d_dataframe_extractor.apply(con, ref_type=ref_type, keep_first=keep_first,
                                                 min_extr_date=min_extr_date, gjahr=gjahr, enable_changes=enable_changes, enable_payments=enable_payments, allowed_act_doc_types=allowed_act_doc_types, allowed_act_changes=allowed_act_changes)
    log = log_converter.apply(dataframe, parameters={"stream_postprocessing": True})
    log = sorting.sort_timestamp(log, "time:timestamp")
    return log


def cli(con):
    print("\n\nO2C XES log extractor\n\n")
    ref_type = input("Insert the central document type of the extraction (default: Invoice): ")
    if not ref_type:
        ref_type = "Invoice"
    ext_type = input("Do you want to extract the document log, or the items log (default: document):")
    if not ext_type:
        ext_type = "document"
    if ext_type == "document":
        keep_first = True
    else:
        keep_first = False
    min_extr_date = input("Insert the minimum extraction date (default: 2020-01-01 00:00:00): ")
    if not min_extr_date:
        min_extr_date = "2020-01-01 00:00:00"
    gjahr = input("Insert the fiscal year (default: 2020):")
    if not gjahr:
        gjahr = "2020"
    log = apply(con, ref_type=ref_type, keep_first=keep_first, min_extr_date=min_extr_date, gjahr=gjahr)
    path = input("Insert the path where the log should be saved (default: o2c.xes):")
    if not path:
        path = "o2c.xes"
    xes_exporter.apply(log, path)



def filter_on_case_size(df, case_id_glue="case:concept:name", min_case_size=2, max_case_size=None):
    """
    Filter a dataframe keeping only traces with at least the specified number of events

    Parameters
    -----------
    df
        Dataframe
    case_id_glue
        Case ID column in the CSV
    min_case_size
        Minimum size of a case
    max_case_size
        Maximum case size

    Returns
    -----------
    df
        Filtered dataframe
    """
    element_group_size = df[case_id_glue].groupby(df[case_id_glue]).transform('size')
    df = df[element_group_size >= min_case_size]
    if max_case_size:
        element_group_size = df[case_id_glue].groupby(df[case_id_glue]).transform('size')
        df = df[element_group_size <= max_case_size]
    return df



from sapextractor.algo.ap_ar import ap_ar_common
from sapextractor.utils.graph_building import build_graph
from sapextractor.utils.filters import case_filter
from sapextractor.utils import constants


def apply(con, gjahr="1997", bukrs=None, ref_type="Goods receipt"):
    bkpf, bseg = ap_ar_common.get_single_dataframes(con, gjahr=gjahr, bukrs=bukrs, filter_columns=True)
    anc_succ = build_graph.get_ancestors_successors(bseg, "event_AUGBL", "event_BELNR", "event_AUGBL_TYPE", "event_BELNR_TYPE", ref_type)
    bkpf = bkpf.merge(anc_succ, left_on="event_BELNR", right_on="node", suffixes=('', '_r'), how="right")
    bkpf = bkpf.reset_index()
    bkpf["event_activity"] = bkpf["event_ONLYACT"] + " (" + bkpf["event_BLART"] + ")"
    ren_cols = {"event_activity": "concept:name", "event_timestamp": "time:timestamp", "event_USNAM": "org:resource"}
    bkpf = bkpf.rename(columns=ren_cols)
    ren_cols = {x: x.split("event_")[-1] for x in bkpf.columns}
    bkpf = bkpf.rename(columns=ren_cols)
    bkpf = bkpf.dropna(subset=["concept:name", "time:timestamp"], how="any")
    bkpf = bkpf.sort_values("time:timestamp")
    bkpf = case_filter.filter_on_case_size(bkpf, "case:concept:name", min_case_size=1, max_case_size=constants.MAX_CASE_SIZE)
    return bkpf


def cli(con):
    print("\n\nAccounting Doc Flow Transactions dataframe extractor\n\n")
    ref_type = input("Insert the central document type of the extraction (default: Goods receipt): ")
    if not ref_type:
        ref_type = "Goods receipt"
    dataframe = apply(con, ref_type=ref_type)
    path = input("Insert the path where the dataframe should be saved (default: doc_flow_transactions.csv): ")
    if not path:
        path = "doc_flow_transactions.csv"
    dataframe.to_csv(path, sep=",", quotechar="\"", index=False)



from sapextractor.database_connection.interface import DatabaseConnection
from sapextractor.utils.string_matching import find_corr
import pandas as pd
from getpass import getpass
from sapextractor.utils import constants
import time


class OracleConnection(DatabaseConnection):
    def __init__(self, hostname="127.0.0.1", port="1521", sid="xe", username="system", password="oracle"):
        import cx_Oracle
        constants.TIMESTAMP_FORMAT = "%Y%m%d %H%M%S"
        self.table_prefix = "SAPSR3."
        self.con = cx_Oracle.connect(username, password, hostname + ":" + str(port) + "/" + str(sid), encoding="UTF-8",
                                     events=True)
        DatabaseConnection.__init__(self)

    def execute_read_sql(self, sql, columns):
        cursor = self.con.cursor()
        cursor.prefetchrows = constants.ORACLE_ARRAYSIZE
        cursor.arraysize = constants.ORACLE_ARRAYSIZE
        print(time.time(), "executing: "+sql)
        cursor.execute(sql)
        stream = []
        df = []
        while True:
            res = cursor.fetchmany(10000)
            if len(res) == 0:
                break
            for row in res:
                el = {}
                for idx, col in enumerate(columns):
                    el[col] = row[idx]
                stream.append(el)
            this_dataframe = pd.DataFrame(stream)
            df.append(this_dataframe)
            stream = None
            stream = []
        if df:
            df = pd.concat(df)
        else:
            df = pd.DataFrame({x: [] for x in columns})
        df.columns = [x.upper() for x in df.columns]
        print(time.time(), "executed: "+sql)
        return df

    def get_list_tables(self):
        cursor = self.con.cursor()
        cursor.execute("SELECT table_name FROM dba_tables WHERE owner = '" + constants.ORACLE_OWNER + "'")
        tables = cursor.fetchall()
        tables = [x[0] for x in tables]
        return tables

    def write_dataframe(self, dataframe, table_name):
        columns = list(dataframe.columns)
        stream = dataframe.to_dict("records")
        stream2 = []
        while stream:
            event = stream.pop(0)
            new_event = []
            for index, c in enumerate(columns):
                val = event[c]
                new_event.append(val)
            stream2.append(tuple(new_event))
        cursor = self.con.cursor()
        try:
            cursor.execute("DROP TABLE "+table_name)
        except:
            pass
        creation_instructions = ["CREATE TABLE "+table_name+" ("]
        for index, c in enumerate(columns):
            creation_instructions.append(c)
            dtype = str(dataframe[c].dtype)
            if dtype == "object":
                creation_instructions.append(" VARCHAR2(3000)")
            elif dtype == "int64":
                creation_instructions.append(" INTEGER")
            elif dtype == "float64":
                creation_instructions.append(" FLOAT")
            if index < len(columns)-1:
                creation_instructions.append(", ")
        creation_instructions.append(")")
        creation_instructions = "".join(creation_instructions)
        cursor.execute(creation_instructions)
        insertmany_instruction = ["INSERT INTO "+table_name+" VALUES ("]
        for index in range(len(columns)):
            insertmany_instruction.append(":%d" % (index+1))
            if index < len(columns)-1:
                insertmany_instruction.append(", ")
        insertmany_instruction.append(")")
        insertmany_instruction = "".join(insertmany_instruction)
        cursor.executemany(insertmany_instruction, stream2)
        self.con.commit()

    def get_columns(self, table_name):
        cursor = self.con.cursor()
        cursor.execute("select column_name from sys.dba_tab_columns where table_name = '%s' and owner = '%s'" % (table_name, constants.ORACLE_OWNER))
        columns = cursor.fetchall()
        columns = [x[0] for x in columns]
        return columns

    def format_table_name(self, table_name):
        return table_name

    def prepare_query(self, table_name, columns):
        table_name = self.format_table_name(table_name)
        table_columns = self.get_columns(table_name)
        new_columns = find_corr.apply(columns, table_columns)
        columns = new_columns
        table_name = constants.ORACLE_OWNER + "." + table_name
        return "SELECT "+",".join(columns)+" FROM "+table_name

    def prepare_and_execute_query(self, table_name, columns, additional_query_part=""):
        query = self.prepare_query(table_name, columns) + additional_query_part
        dataframe = self.execute_read_sql(query, columns)
        dataframe.columns = columns
        return dataframe


def apply(hostname="127.0.0.1", port="1521", sid="xe", username="system", password="oracle"):
    return OracleConnection(hostname=hostname, port=port, sid=sid, username=username, password=password)


def cli():
    print("\n\n")
    print("== Connection to an Oracle database == \n\n")
    hostname = str(input("Insert the hostname (default: 127.0.0.1):"))
    if not hostname:
        hostname = "127.0.0.1"
    port = str(input("Insert the port (default: 1521):"))
    if not port:
        port = "1521"
    sid = str(input("Insert the SID (default: xe):"))
    if not sid:
        sid = "xe"
    username = str(input("Insert the username of a DB user (default: system):"))
    if not username:
        username = "system"
    password = str(getpass("Insert the password of the DB user (default: oracle):"))
    if not password:
        password = "oracle"
    return apply(hostname=hostname, port=port, sid=sid, username=username, password=password)
